﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;

namespace SilverlightApplication14
{
	// 結果を出力するデリゲート
	delegate void OutPutDelegate(String str);

	/// <summary>
	/// 海辺の美女問題を遺伝的アルゴリズムで解く
	/// <summary>
	public class GA_SeaShore
	{
        const int GENESIZE = 20;	// 女の子との出会い回数 = 遺伝子の長さ
        const int POOLSIZE = 60;	// 遺伝子プールのサイズ
		const int CO_POOLSIZE = POOLSIZE * 2;	// 子のプールサイズ
		const int M_PROB  = 20;		// 突然変異確率(%)
		const int M_PROB2 = 30;		// ビット変異確率(%)
		const int EVALUATE_TIMES = GENESIZE * 8;	// １回の評価での試行回数
        const int GENERATION = 100;	// 世代数 = １ボタンでの試行回数
		private int gene_count;	// 世代数のカウンター 

		protected Gene[] Parent = new Gene[POOLSIZE];
		protected Gene[] Son = new Gene[CO_POOLSIZE];
		protected Gene Elite;

	//	private Random Rnd; // 乱数系列は全体で１つ
		private jp.takel.PseudoRandom.MersenneTwister Rnd;

		// 評価用配列
		// 偶然に出現する女の子のシーケンス
		// 数字が小さい方が順位の高い女の子とする
		private int[] Girls = new int[GENESIZE];

		/// <summary>
		/// コンストラクター
		/// <summary>
		public GA_SeaShore()
		{
			uint seed = (uint)Environment.TickCount;
			Rnd = new jp.takel.PseudoRandom.MersenneTwister(seed);
			init();
        }

		// 結果を出力するデリゲート
		private OutPutDelegate _output;

		public void setOutput(MainPage main)
		{
			_output = new OutPutDelegate(main.appendText);
		}

		/// <summary>
        /// 初期化
		/// <summary>
        public void init()
        {
            int idx;

            // 評価用配列を初期化
            for (idx = 0; idx < GENESIZE; idx++)
            {
                Girls[idx] = (idx + 1);   // １始まりのシーケンシャルな番号を入れておく
            }

            // 親プールを初期化
            for (idx = 0; idx < POOLSIZE; idx++)
            {
                Parent[idx] = new Gene(GENESIZE, Rnd);
                evaluate( ref Parent[idx] );
            }
            // 子プールを初期化
			for (idx = 0; idx < CO_POOLSIZE; idx++)
            {
                Son[idx] = new Gene(GENESIZE, Rnd);
            //  evaluate(ref Parent[idx]);  // 不要
            }
            // エリートの初期化
            Elite = new Gene(GENESIZE, Rnd);
            Elite.point = -1.0;
            keepElite();

			gene_count = 0;	// 世代カウンターリセット
		}

		/// <summary>
        /// 実行
		/// </summary>
		public void run()
		{
			for (uint idx = 0; idx < GENERATION; idx++)
			{
				crossing();
				mutation();
				selection();
				keepElite();

				// 結果を出力
				_output( (++gene_count).ToString() + " : "
					+ (GENESIZE - Elite.point).ToString() + " : "
					+  Elite.print() +"\n" );
			}
        }

		/// <summary>
        /// 交差
		/// <summary>
		protected void crossing()
        {
            int i, j;
            int f = 0, m = 0;
            int cp;

			for (i = 0; i < CO_POOLSIZE; i++)
            {
                while (f == m)
                {
                    f = Rnd.Next(POOLSIZE);
                    m = Rnd.Next(POOLSIZE);
                }
                cp = Rnd.Next(GENESIZE);
                for (j = 0; j < GENESIZE; j++)
                {
                    if (j < cp)
                    {
                        Son[i].genotype[j] = Parent[f].genotype[j];
                    }
                    else
                    {
                        Son[i].genotype[j] = Parent[m].genotype[j];
                    }
                }
                evaluate( ref Son[i]);
            }
        }

		/// <summary>
        /// 突然変異
		/// <summary>
		protected void mutation()
        {
			int	i, j;

			for (i = 0; i < CO_POOLSIZE; i++)
            {
				if( Rnd.Next(100) < M_PROB )
				{
					for( j = 0; j < GENESIZE; j++ )
					{
						if( Rnd.Next(100) <= M_PROB2 )
                        {
							Son[i].genotype[j] = Rnd.Next(GENESIZE);
						}
			        }
					evaluate( ref Son[i] );
		        }
			}
		}

		/// <summary>
        /// 淘汰
		/// <summary>
        protected void selection()
        {
            double point = 0.0;
            double prob, p = 0.0, popprob;
            int i, j;

            // ルーレットの作成
			for (i = 0; i < CO_POOLSIZE; i++)
            {
                point += Son[i].point;
            }
			for (i = 0; i < CO_POOLSIZE; i++)
            {
                Son[i].prob = Son[i].point / point;
            }
            popprob = 1.0;

            // ルーレットを回す
            for (i = 0; i < POOLSIZE; i++)
            {
                prob = Rnd.NextDouble() * popprob;
				for (j = 0; j < CO_POOLSIZE; j++)
                {
                    if (Son[j].prob != -1.0)
                    {
                        p += Son[j].prob;
                    }
                    if (p >= prob)
                    {
                        // 選ばれた遺伝子をルーレットから除く
                        popprob -= Son[j].prob;
                        Son[j].prob = -1.0;
                        // 選ばれた遺伝子を親プールに格納
                        Parent[i] = Son[j].Copy();
                        break;
                    }
                }
            }
        }

		/// <summary>
        /// エリート保存
		/// </summary>
        protected void keepElite()
        {
            int idx;

            for (idx = 0; idx < POOLSIZE; idx++)
			{
                if (Elite.point < Parent[idx].point)
				{
                    Elite = Parent[idx].Copy();
					// 一度エリートになった個体は淘汰からは隔離される
				}
			}

			// 適当な1個体をエリートに書き換える
			// エリートの血を絶やさないために
			idx = Rnd.Next(POOLSIZE);
			Parent[idx] = Elite.Copy();
        }

		/// <summary>
        /// 評価
		/// </summary>
		protected void evaluate( ref Gene g )
		{
			int sum_level = 0;

			// 遺伝子の表現型を作る
			g.invoke();

            // 何度か試してみる
			for (int trial = 0; trial < EVALUATE_TIMES; trial++)
			{
				_shuffle( ref Girls );

                // 出会った女の子の記憶リスト
				List<int> memories = new List<int>();
				// 先頭に「高嶺の花」、ラストに「最低」を入れておく
                // Girlsの中身は 1 ～ GENESIZE なので、それより小さい数と、大きい数
				memories.Add( 0 );
				memories.Add( GENESIZE + 1 );
				
				// 評価用配列に対して、頭から順番にトライする
				for (int idx = 0; idx < GENESIZE; idx++)
				{
                    int a_girl = Girls[idx];	// 今回出会った女の子の順位
					int level = g.phenotype[idx];	// 現在の許容レベル
					if (memories.Count <= level		// 今まで出会った数よりも許容レベルが低い
					 || a_girl < memories[level] )	// 今回出会った子の方が上
									// リスト末尾の「最低」より上、も含んでいる
					{
						sum_level += a_girl;
						break;
					}
					// 声をかけなかったときは、今回出会った子をメモリーに加える
					_insertMemory(ref memories, a_girl);
				}
			}

            // (全数 - 得られた平均順位)を評価値とする
			// ・・・順位数が小さいほど、ポイントを高くしたかったので。
			g.point = GENESIZE - (double)sum_level / EVALUATE_TIMES;
		}

		/// <summary>
        /// a_girl を memories の順序を崩さない位置に挿入する
        /// memories は降順に並んでいる
		/// <summary>
		private void _insertMemory(ref List<int> memories, int a_girl)
		{
            int idx = 0;
			
			foreach (int scan in memories)
			{
				if (a_girl < scan)
				{
					break;
				}
                idx++;
			}
            memories.Insert(idx, a_girl);
		}

		/// <summary>
        /// 評価用配列をかき混ぜる
		/// </summary>
		private void _shuffle( ref int[] girls )
		{
            // GENESIZE 回の入れ替えを行う
			for (int idx = 0; idx < GENESIZE; idx++)
			{
				int a = Rnd.Next(GENESIZE);
                int b = Rnd.Next(GENESIZE);
                if (a == b)
				{
					continue;
				}
				_swap(ref girls[a], ref girls[b]);
			}
		}

		/// <summary>
		/// a と b の中身を入れ替える。
		/// </summary>
		private void _swap<T>(ref T a, ref T b)
		{
            T c = a; a = b; b = c;
		}
	}
}
